"""Tests for the session manager."""

from functools import partial
from unittest import TestCase

from tornado import gen, web
from tornado.ioloop import IOLoop

from ..sessionmanager import SessionManager
from jupyter_server.services.kernels.kernelmanager import MappingKernelManager
from jupyter_server.services.contents.manager import ContentsManager
from jupyter_server._tz import utcnow, isoformat


class DummyKernel(object):
    def __init__(self, kernel_name='python'):
        self.kernel_name = kernel_name

dummy_date = utcnow()
dummy_date_s = isoformat(dummy_date)


class DummyMKM(MappingKernelManager):
    """MappingKernelManager interface that doesn't start kernels, for testing"""

    def __init__(self, *args, **kwargs):
        super(DummyMKM, self).__init__(*args, **kwargs)
        self.id_letters = iter(u'ABCDEFGHIJK')

    def _new_id(self):
        return next(self.id_letters)

    def start_kernel(self, kernel_id=None, path=None, kernel_name='python', **kwargs):
        kernel_id = kernel_id or self._new_id()
        k = self._kernels[kernel_id] = DummyKernel(kernel_name=kernel_name)
        self._kernel_connections[kernel_id] = 0
        k.last_activity = dummy_date
        k.execution_state = 'idle'
        return kernel_id

    def shutdown_kernel(self, kernel_id, now=False):
        del self._kernels[kernel_id]


class TestSessionManager(TestCase):

    def setUp(self):
        self.sm = SessionManager(
            kernel_manager=DummyMKM(),
            contents_manager=ContentsManager(),
        )
        self.loop = IOLoop()
        self.addCleanup(partial(self.loop.close, all_fds=True))

    def create_sessions(self, *kwarg_list):
        @gen.coroutine
        def co_add():
            sessions = []
            for kwargs in kwarg_list:
                kwargs.setdefault('type', 'notebook')
                session = yield self.sm.create_session(**kwargs)
                sessions.append(session)
            raise gen.Return(sessions)
        return self.loop.run_sync(co_add)

    def create_session(self, **kwargs):
        return self.create_sessions(kwargs)[0]

    def test_get_session(self):
        sm = self.sm
        session_id = self.create_session(path='/path/to/test.ipynb', kernel_name='bar')['id']
        model = sm.get_session(session_id=session_id)
        expected = {'id':session_id,
                    'path': u'/path/to/test.ipynb',
                    'notebook': {'path': u'/path/to/test.ipynb', 'name': None},
                    'type': 'notebook',
                    'name': None,
                    'kernel': {
                        'id': 'A',
                        'name': 'bar',
                        'connections': 0,
                        'last_activity': dummy_date_s,
                        'execution_state': 'idle',
                    }}
        self.assertEqual(model, expected)

    def test_bad_get_session(self):
        # Should raise error if a bad key is passed to the database.
        sm = self.sm
        session_id = self.create_session(path='/path/to/test.ipynb',
                                       kernel_name='foo')['id']
        self.assertRaises(TypeError, sm.get_session, bad_id=session_id) # Bad keyword

    def test_get_session_dead_kernel(self):
        sm = self.sm
        session = self.create_session(path='/path/to/1/test1.ipynb', kernel_name='python')
        # kill the kernel
        sm.kernel_manager.shutdown_kernel(session['kernel']['id'])
        with self.assertRaises(KeyError):
            sm.get_session(session_id=session['id'])
        # no sessions left
        listed = sm.list_sessions()
        self.assertEqual(listed, [])

    def test_list_sessions(self):
        sm = self.sm
        sessions = self.create_sessions(
            dict(path='/path/to/1/test1.ipynb', kernel_name='python'),
            dict(path='/path/to/2/test2.py', type='file', kernel_name='python'),
            dict(path='/path/to/3', name='foo', type='console', kernel_name='python'),
        )

        sessions = sm.list_sessions()
        expected = [
            {
                'id':sessions[0]['id'],
                'path': u'/path/to/1/test1.ipynb',
                'type': 'notebook',
                'notebook': {'path': u'/path/to/1/test1.ipynb', 'name': None},
                'name': None,
                'kernel': {
                    'id': 'A',
                    'name':'python',
                    'connections': 0,
                    'last_activity': dummy_date_s,
                    'execution_state': 'idle',
                }
            }, {
                'id':sessions[1]['id'],
                'path': u'/path/to/2/test2.py',
                'type': 'file',
                'name': None,
                'kernel': {
                    'id': 'B',
                    'name':'python',
                    'connections': 0,
                    'last_activity': dummy_date_s,
                    'execution_state': 'idle',
                }
            }, {
                'id':sessions[2]['id'],
                'path': u'/path/to/3',
                'type': 'console',
                'name': 'foo',
                'kernel': {
                    'id': 'C',
                    'name':'python',
                    'connections': 0,
                    'last_activity': dummy_date_s,
                    'execution_state': 'idle',
                }
            }
        ]
        self.assertEqual(sessions, expected)

    def test_list_sessions_dead_kernel(self):
        sm = self.sm
        sessions = self.create_sessions(
            dict(path='/path/to/1/test1.ipynb', kernel_name='python'),
            dict(path='/path/to/2/test2.ipynb', kernel_name='python'),
        )
        # kill one of the kernels
        sm.kernel_manager.shutdown_kernel(sessions[0]['kernel']['id'])
        listed = sm.list_sessions()
        expected = [
            {
                'id': sessions[1]['id'],
                'path': u'/path/to/2/test2.ipynb',
                'type': 'notebook',
                'name': None,
                'notebook': {'path': u'/path/to/2/test2.ipynb', 'name': None},
                'kernel': {
                    'id': 'B',
                    'name':'python',
                    'connections': 0,
                    'last_activity': dummy_date_s,
                    'execution_state': 'idle',
                }
            }
        ]
        self.assertEqual(listed, expected)

    def test_update_session(self):
        sm = self.sm
        session_id = self.create_session(path='/path/to/test.ipynb',
                                       kernel_name='julia')['id']
        sm.update_session(session_id, path='/path/to/new_name.ipynb')
        model = sm.get_session(session_id=session_id)
        expected = {'id':session_id,
                    'path': u'/path/to/new_name.ipynb',
                    'type': 'notebook',
                    'name': None,
                    'notebook': {'path': u'/path/to/new_name.ipynb', 'name': None},
                    'kernel': {
                        'id': 'A',
                        'name':'julia',
                        'connections': 0,
                        'last_activity': dummy_date_s,
                        'execution_state': 'idle',
                    }
        }
        self.assertEqual(model, expected)

    def test_bad_update_session(self):
        # try to update a session with a bad keyword ~ raise error
        sm = self.sm
        session_id = self.create_session(path='/path/to/test.ipynb',
                                       kernel_name='ir')['id']
        self.assertRaises(TypeError, sm.update_session, session_id=session_id, bad_kw='test.ipynb') # Bad keyword

    def test_delete_session(self):
        sm = self.sm
        sessions = self.create_sessions(
            dict(path='/path/to/1/test1.ipynb', kernel_name='python'),
            dict(path='/path/to/2/test2.ipynb', kernel_name='python'),
            dict(path='/path/to/3', name='foo', type='console', kernel_name='python'),
        )
        sm.delete_session(sessions[1]['id'])
        new_sessions = sm.list_sessions()
        expected = [{
                'id': sessions[0]['id'],
                'path': u'/path/to/1/test1.ipynb',
                'type': 'notebook',
                'name': None,
                'notebook': {'path': u'/path/to/1/test1.ipynb', 'name': None},
                'kernel': {
                    'id': 'A',
                    'name':'python',
                    'connections': 0,
                    'last_activity': dummy_date_s,
                    'execution_state': 'idle',
                }
            }, {
                'id': sessions[2]['id'],
                'type': 'console',
                'path': u'/path/to/3',
                'name': 'foo',
                'kernel': {
                    'id': 'C',
                    'name':'python',
                    'connections': 0,
                    'last_activity': dummy_date_s,
                    'execution_state': 'idle',
                }
            }
        ]
        self.assertEqual(new_sessions, expected)

    def test_bad_delete_session(self):
        # try to delete a session that doesn't exist ~ raise error
        sm = self.sm
        self.create_session(path='/path/to/test.ipynb', kernel_name='python')
        with self.assertRaises(TypeError):
            self.loop.run_sync(lambda : sm.delete_session(bad_kwarg='23424')) # Bad keyword
        with self.assertRaises(web.HTTPError):
            self.loop.run_sync(lambda : sm.delete_session(session_id='23424')) # nonexistent




import os
import tempfile
import nose.tools as nt

from ...extension.application import ExtensionApp
from ...serverapp import ServerApp


class MockExtension(ExtensionApp):
    extension_name = 'mock'        


class ExtensionTestingMixin:

    port = 12341

    def server_kwargs(self):
        return dict(
            port=self.port,
            port_retries=0,
            open_browser=False,
            allow_root=True
        )

    def tearDown(self):
        #self.serverapp.stop()
        self.serverapp.http_server.stop()
        self.serverapp.clear_instance()


class TestExtensionAppInitialize(ExtensionTestingMixin):

    def setUp(self):
        self.serverapp = ServerApp(**self.server_kwargs())
        self.serverapp.init_signal = lambda : None
        self.serverapp.init_terminals = lambda : None
        # clear log handlers and propagate to root for nose to capture it
        # needs to be redone after initialize, which reconfigures logging
        self.serverapp.log.propagate = True
        self.serverapp.log.handlers = []
        self.serverapp.initialize(argv=[])

    def test_instance_creation(self):
        self.extension = MockExtension()
        extension = self.extension
        nt.assert_equal(extension.static_paths, [])
        nt.assert_equal(extension.template_paths, [])
        nt.assert_equal(extension.settings, {})
        nt.assert_equal(extension.handlers, [])

    def test_instance_creation_with_arg(self):
        self.extension = MockExtension(static_paths=['test'])
        nt.assert_in('test', self.extension.static_paths)

    def test_initialize(self):
        self.extension = MockExtension()
        self.extension.initialize(self.serverapp)
        nt.assert_true(isinstance(self.extension.serverapp, ServerApp))


class TestExtensionServerInitialize(ExtensionTestingMixin):

    def test_initialize_server(self):
        self.extension = MockExtension()
        self.serverapp = self.extension.initialize_server(**self.server_kwargs())
        nt.assert_true(isinstance(self.serverapp, ServerApp))

    def test_initialize_server_argv(self):
        argv = [
            '--ServerApp.tornado_settings={"test":"hello world"}'
        ]
        self.extension = MockExtension()
        self.serverapp = self.extension.initialize_server(
            argv=argv, **self.server_kwargs())

        nt.assert_in("test", self.serverapp.tornado_settings)

    def test_initialize_mixed_argv(self):
        argv = [
            '--ServerApp.tornado_settings={"test":"hello world"}',
            '--MockExtension.static_paths=["test"]'
        ]
        # Initialize server
        self.extension = MockExtension()
        self.serverapp = self.extension.initialize_server(
            argv=argv,
            **self.server_kwargs()
        )
        self.extension.initialize(self.serverapp, argv=argv)
        nt.assert_in("test", self.serverapp.tornado_settings)
        nt.assert_in("test", self.extension.static_paths)




"""Test ServerApp"""

import getpass
import logging
import os
import re
import signal
from subprocess import Popen, PIPE, STDOUT
import sys
from tempfile import NamedTemporaryFile

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch # py2

import nose.tools as nt

from ipython_genutils.tempdir import TemporaryDirectory

from traitlets.tests.utils import check_help_all_output
from traitlets import TraitError

from jupyter_core.application import NoStart
from jupyter_server import serverapp, __version__
from jupyter_server.auth.security import passwd_check

ServerApp = serverapp.ServerApp


def test_help_output():
    """jupyter server --help-all works"""
    check_help_all_output('jupyter_server')

def test_server_info_file():
    td = TemporaryDirectory()
    svapp = ServerApp(runtime_dir=td.name, log=logging.getLogger())

    def get_servers():
        return list(serverapp.list_running_servers(svapp.runtime_dir))

    svapp.initialize(argv=[])
    svapp.write_server_info_file()
    servers = get_servers()
    nt.assert_equal(len(servers), 1)
    nt.assert_equal(servers[0]['port'], svapp.port)
    nt.assert_equal(servers[0]['url'], svapp.connection_url)
    svapp.remove_server_info_file()
    nt.assert_equal(get_servers(), [])

    # The ENOENT error should be silenced.
    svapp.remove_server_info_file()

def test_root_dir():
    with TemporaryDirectory() as td:
        app = ServerApp(root_dir=td)
        nt.assert_equal(app.root_dir, td)

def test_no_create_root_dir():
    with TemporaryDirectory() as td:
        rootdir = os.path.join(td, 'notebooks')
        app = ServerApp()
        with nt.assert_raises(TraitError):
            app.root_dir = rootdir

def test_missing_root_dir():
    with TemporaryDirectory() as td:
        rootdir = os.path.join(td, 'root', 'dir', 'is', 'missing')
        app = ServerApp()
        with nt.assert_raises(TraitError):
            app.root_dir = rootdir

def test_invalid_root_dir():
    with NamedTemporaryFile() as tf:
        app = ServerApp()
        with nt.assert_raises(TraitError):
            app.root_dir = tf

def test_root_dir_with_slash():
    with TemporaryDirectory(suffix="_slash" + os.sep) as td:
        app = ServerApp(root_dir=td)
        nt.assert_false(app.root_dir.endswith(os.sep))

def test_root_dir_root():
    root = os.path.abspath(os.sep) # gets the right value on Windows, Posix
    app = ServerApp(root_dir=root)
    nt.assert_equal(app.root_dir, root)

def test_generate_config():
    with TemporaryDirectory() as td:
        app = ServerApp(config_dir=td)
        app.initialize(['--generate-config', '--allow-root'])
        with nt.assert_raises(NoStart):
            app.start()
        assert os.path.exists(os.path.join(td, 'jupyter_server_config.py'))

#test if the version testin function works
def test_pep440_version():

    for version in [
        '4.1.0.b1',
        '4.1.b1',
        '4.2',
        'X.y.z',
        '1.2.3.dev1.post2',
        ]:
        def loc():
            with nt.assert_raises(ValueError):
                raise_on_bad_version(version)
        yield loc

    for version in [
        '4.1.1',
        '4.2.1b3',
        ]:

        yield (raise_on_bad_version, version)

pep440re = re.compile('^(\d+)\.(\d+)\.(\d+((a|b|rc)\d+)?)(\.post\d+)?(\.dev\d*)?$')

def raise_on_bad_version(version):
    if not pep440re.match(version):
        raise ValueError("Versions String does apparently not match Pep 440 specification, "
                         "which might lead to sdist and wheel being seen as 2 different release. "
                         "E.g: do not use dots for beta/alpha/rc markers.")

def test_current_version():
    raise_on_bad_version(__version__)

def test_server_password():
    password = 'secret'
    with TemporaryDirectory() as td:
        with patch.dict('os.environ', {
            'JUPYTER_CONFIG_DIR': td,
        }), patch.object(getpass, 'getpass', return_value=password):
            app = serverapp.JupyterPasswordApp(log_level=logging.ERROR)
            app.initialize([])
            app.start()
            sv = ServerApp()
            sv.load_config_file()
            nt.assert_not_equal(sv.password, '')
            passwd_check(sv.password, password)

class TestingStopApp(serverapp.JupyterServerStopApp):
    """For testing the logic of JupyterServerStopApp."""
    def __init__(self, **kwargs):
        super(TestingStopApp, self).__init__(**kwargs)
        self.servers_shut_down = []

    def shutdown_server(self, server):
        self.servers_shut_down.append(server)
        return True

def test_server_stop():
    def list_running_servers(runtime_dir):
        for port in range(100, 110):
            yield {
                'pid': 1000 + port,
                'port': port,
                'base_url': '/',
                'hostname': 'localhost',
                'root_dir': '/',
                'secure': False,
                'token': '',
                'password': False,
                'url': 'http://localhost:%i' % port,
            }

    mock_servers = patch('jupyter_server.serverapp.list_running_servers', list_running_servers)

    # test stop with a match
    with mock_servers:
        app = TestingStopApp()
        app.initialize(['105'])
        app.start()
    nt.assert_equal(len(app.servers_shut_down), 1)
    nt.assert_equal(app.servers_shut_down[0]['port'], 105)

    # test no match
    with mock_servers, patch('os.kill') as os_kill:
        app = TestingStopApp()
        app.initialize(['999'])
        with nt.assert_raises(SystemExit) as exc:
            app.start()
        nt.assert_equal(exc.exception.code, 1)
    nt.assert_equal(len(app.servers_shut_down), 0)



"""Test the basic /api endpoints"""

import requests

from jupyter_server._tz import isoformat
from jupyter_server.utils import url_path_join
from jupyter_server.tests.launchserver import ServerTestBase


class KernelAPITest(ServerTestBase):
    """Test the kernels web service API"""

    def _req(self, verb, path, **kwargs):
        r = self.request(verb, url_path_join('api', path))
        r.raise_for_status()
        return r

    def get(self, path, **kwargs):
        return self._req('GET', path)

    def test_get_spec(self):
        r = self.get('spec.yaml')
        assert r.text

    def test_get_status(self):
        r = self.get('status')
        data = r.json()
        assert data['connections'] == 0
        assert data['kernels'] == 0
        assert data['last_activity'].endswith('Z')
        assert data['started'].endswith('Z')
        assert data['started'] == isoformat(self.server.web_app.settings['started'])



import sys

from traitlets import (
    Unicode, 
    List, 
    Dict, 
    default, 
    validate
)
from traitlets.config import Config

from jupyter_core.application import JupyterApp

from jupyter_server.serverapp import ServerApp, aliases, flags
from jupyter_server.transutils import _
from jupyter_server.utils import url_path_join


# Remove alias for nested classes in ServerApp.
# Nested classes are not allowed in ExtensionApp.
try:
    aliases.pop('transport')
except KeyError:
    pass


def _preparse_command_line(Application):
    """Looks for 'help', 'version', and 'generate-config; commands 
    in command line. If found, raises the help and version of 
    current Application.

    This is useful for traitlets applications that have to parse
    the command line multiple times, but want to control when
    when 'help' and 'version' is raised.
    """
    # Arguments after a '--' argument are for the script IPython may be
    # about to run, not IPython iteslf. For arguments parsed here (help and
    # version), we want to only search the arguments up to the first
    # occurrence of '--', which we're calling interpreted_argv.
    try:
        interpreted_argv = sys.argv[:sys.argv.index('--')]
    except ValueError:
        interpreted_argv = sys.argv

    # Catch any help calls.
    if any(x in interpreted_argv for x in ('-h', '--help-all', '--help')):
        app = Application()
        app.print_help('--help-all' in interpreted_argv)
        app.exit(0)

    # Catch version commands
    if '--version' in interpreted_argv or '-V' in interpreted_argv:
        app = Application()
        app.print_version()
        app.exit(0)

    # Catch generate-config commands.
    if '--generate-config' in interpreted_argv:
        app = Application()
        app.write_default_config()
        app.exit(0)


class ExtensionApp(JupyterApp):
    """Base class for configurable Jupyter Server Extension Applications.

    ExtensionApp subclasses can be initialized two ways:
    1. Extension is listed as a jpserver_extension, and ServerApp calls 
        its load_jupyter_server_extension classmethod. This is the 
        classic way of loading a server extension.
    2. Extension is launched directly by calling its `launch_instance`
        class method. This method can be set as a entry_point in 
        the extensions setup.py
    """
    # Name of the extension
    extension_name = Unicode(
        "",
        help="Name of extension."
    )

    @default("extension_name")
    def _default_extension_name(self):
        raise Exception("The extension must be given a `name`.")

    @validate("extension_name")
    def _valid_extension_name(self, obj, value):
        if isinstance(name, str):
            # Validate that extension_name doesn't contain any invalid characters.
            for char in [' ', '.', '+', '/']:
                self.error(obj, value)
            return value
        self.error(obj, value) 

    # Extension can configure the ServerApp from the command-line
    classes = [
        ServerApp,
    ]

    aliases = aliases
    flags = flags

    @property
    def static_url_prefix(self):
        return "/static/{extension_name}/".format(
            extension_name=self.extension_name)

    static_paths = List(Unicode(),
        help="""paths to search for serving static files.
        
        This allows adding javascript/css to be available from the notebook server machine,
        or overriding individual files in the IPython
        """
    ).tag(config=True)

    template_paths = List(Unicode(), 
        help=_("""Paths to search for serving jinja templates.

        Can be used to override templates from notebook.templates.""")
    ).tag(config=True)

    settings = Dict(
        help=_("""Settings that will passed to the server.""")
    ).tag(config=True)

    handlers = List(
        help=_("""Handlers appended to the server.""")
    ).tag(config=True)

    default_url = Unicode('/', config=True,
        help=_("The default URL to redirect to from `/`")
    )

    def initialize_settings(self):
        """Override this method to add handling of settings."""
        pass

    def initialize_handlers(self):
        """Override this method to append handlers to a Jupyter Server."""
        pass

    def initialize_templates(self):
        """Override this method to add handling of template files."""
        pass

    def _prepare_config(self):
        """Builds a Config object from the extension's traits and passes
        the object to the webapp's settings as `<extension_name>_config`.  
        """
        traits = self.class_own_traits().keys()
        self.config = Config({t: getattr(self, t) for t in traits})
        self.settings['{}_config'.format(self.extension_name)] = self.config

    def _prepare_settings(self):
        # Make webapp settings accessible to initialize_settings method
        webapp = self.serverapp.web_app
        self.settings.update(**webapp.settings)

        # Add static and template paths to settings.
        self.settings.update({
            "{}_static_paths".format(self.extension_name): self.static_paths,
        })

        # Get setting defined by subclass using initialize_settings method.
        self.initialize_settings()

        # Update server settings with extension settings.
        webapp.settings.update(**self.settings)

    def _prepare_handlers(self):
        webapp = self.serverapp.web_app

        # Get handlers defined by extension subclass.
        self.initialize_handlers()

        # prepend base_url onto the patterns that we match
        new_handlers = []
        for handler_items in self.handlers:
            # Build url pattern including base_url
            pattern = url_path_join(webapp.settings['base_url'], handler_items[0])
            handler = handler_items[1]
            
            # Get handler kwargs, if given
            kwargs = {}
            try: 
                kwargs.update(handler_items[2])
            except IndexError:
                pass
            kwargs['extension_name'] = self.extension_name

            new_handler = (pattern, handler, kwargs)
            new_handlers.append(new_handler)

        # Add static endpoint for this extension, if static paths are given.
        if len(self.static_paths) > 0:
            # Append the extension's static directory to server handlers.
            static_url = url_path_join("/static", self.extension_name, "(.*)")
            
            # Construct handler.
            handler = (
                static_url, 
                webapp.settings['static_handler_class'], 
                {'path': self.static_paths}
            )
            new_handlers.append(handler)

        webapp.add_handlers('.*$', new_handlers)

    def _prepare_templates(self):
        # Add templates to web app settings if extension has templates.
        if len(self.template_paths) > 0:
            self.settings.update({
                "{}_template_paths".format(self.extension_name): self.template_paths
            })
        self.initialize_templates()

    @staticmethod
    def initialize_server(argv=[], **kwargs):
        """Get an instance of the Jupyter Server."""
        # Get a jupyter server instance
        serverapp = ServerApp(**kwargs)
        # Initialize ServerApp config.
        # Parses the command line looking for 
        # ServerApp configuration.
        serverapp.initialize(argv=argv)
        return serverapp

    def initialize(self, serverapp, argv=[]):
        """Initialize the extension app.
        
        This method:
        - Loads the extension's config from file
        - Updates the extension's config from argv
        - Initializes templates environment
        - Passes settings to webapp
        - Appends handlers to webapp.
        """
        # Initialize the extension application
        super(ExtensionApp, self).initialize(argv=argv)
        self.serverapp = serverapp

        # Intialize config, settings, templates, and handlers.
        self._prepare_config()
        self._prepare_templates()
        self._prepare_settings()
        self._prepare_handlers()

    def listen(self):
        """Extension doesn't listen for anything. IOLoop lives
        in the attached server.
        """
        pass

    def start(self):
        """Extension has nothing to `start`. See `start_server`
        for starting the jupyter server application.
        """
        pass

    def start_server(self, **kwargs):
        """Start the Jupyter server.
        
        Server should be started after extension is initialized.
        """
        # Start the server.
        self.serverapp.start(**kwargs)

    @classmethod
    def load_jupyter_server_extension(cls, serverapp, argv=[], **kwargs):
        """Initialize and configure this extension, then add the extension's
        settings and handlers to the server's web application.
        """
        # Configure and initialize extension.
        extension = cls()
        extension.initialize(serverapp, argv=argv)

        return extension

    @classmethod
    def _prepare_launch(cls, serverapp, argv=[], **kwargs):
        """Prepare the extension application for launch by 
        configuring the server and the extension from argv.
        Does not start the ioloop.
        """
        # Load the extension
        extension = cls.load_jupyter_server_extension(serverapp, argv=argv, **kwargs)
        
        # Start the browser at this extensions default_url, unless user
        # configures ServerApp.default_url on command line.
        try:
            server_config = extension.config['ServerApp']
            if 'default_url' not in server_config:
                serverapp.default_url = extension.default_url
        except KeyError: 
            pass

        return extension

    @classmethod
    def launch_instance(cls, argv=None, **kwargs):
        """Launch the extension like an application. Initializes+configs a stock server 
        and appends the extension to the server. Then starts the server and routes to
        extension's landing page.
        """
        # Check for help, version, and generate-config arguments
        # before initializing server to make sure these
        # arguments trigger actions from the extension not the server.
        _preparse_command_line(cls)
        # Handle arguments.
        if argv is not None:
            args = sys.argv[1:]  # slice out extension config.
        else:
            args = []
        
        # Get a jupyter server instance.
        serverapp = cls.initialize_server(argv=args)
        extension = cls._prepare_launch(serverapp, argv=args, **kwargs)
        # Start the ioloop.
        extension.start_server()

